module Degenz.Store

open System.Threading.Tasks
open DSharpPlus.Entities
open DSharpPlus
open DSharpPlus.EventArgs
open DSharpPlus.SlashCommands
open Degenz
open Degenz.Messaging

let checkHasSufficientFunds (item : Item) player =
    if player.Bank - item.Price >= 0<GBT>
       then Ok player
       else Error $"You do not have sufficient funds to buy this item! Current balance: {player.Bank} GBT"

let checkAlreadyOwnsItem (item : Item) player =
    if player.Inventory |> Array.exists (fun w -> item.Id = w.Id)
        then Error $"You already own {item.Name}!"
        else Ok player

let checkSoldItemAlready (item : Item) player =
    if player.Inventory |> Array.exists (fun w -> item.Id = w.Id)
        then Ok player
        else Error $"{item.Name} not found in your arsenal! Looks like you sold it already."

let checkHasItemsInArsenal itemType player =
    if player.Inventory |> Array.filter (fun i -> i.Type = itemType ) |> Array.length > 0
        then Ok player
        else Error $"You currently have no {itemType}s in your arsenal to sell!"

let buy itemType (ctx : IDiscordContext) =
    Game.executePlayerAction ctx (fun player -> async {
        let itemStore = Embeds.getBuyItemsEmbed player itemType Armory.battleItems
        do! ctx.FollowUp itemStore |> Async.AwaitTask
    })

let sell itemType (ctx : IDiscordContext) =
    Game.executePlayerAction ctx (fun player -> async {
        match checkHasItemsInArsenal itemType player with
        | Ok _ -> let itemStore = Embeds.getSellItemsEmbed itemType player
                  do! ctx.FollowUp(itemStore) |> Async.AwaitTask
        | Error e -> do! sendFollowUpMessage ctx e
    })

// TODO: When you buy a shield, prompt the user to activate it
let handleBuyItem (ctx : IDiscordContext) itemId =
    Game.executePlayerAction ctx (fun player -> async {
        let item = Armory.battleItems |> Array.find (fun w -> w.Id = itemId)
        do! player
            |> checkHasSufficientFunds item
            >>= checkAlreadyOwnsItem item
            |> handleResultWithResponse ctx (fun player -> async {
                let newBalance = player.Bank - item.Price
                let p = { player with Bank = newBalance ; Inventory = Array.append [| item |] player.Inventory }
                do! DbService.updatePlayer p
                do! sendFollowUpMessage ctx $"Successfully purchased {item.Name}! You now have {newBalance} 💰$GBT remaining"
            })
    })

let handleSell (ctx : IDiscordContext) itemId =
    Game.executePlayerAction ctx (fun player -> async {
        let item = Armory.getItem itemId
        do! player
            |> checkSoldItemAlready item
            |> handleResultWithResponse ctx (fun player -> async {
                let updatedPlayer = {
                    player with
                        Bank = player.Bank + item.Price
                        Inventory = player.Inventory |> Array.filter (fun i -> i.Id <> itemId)
                        Events =
                            if item.Type = ItemType.Shield
                                then player.Events |> Array.filter (fun a -> a.ItemId <> itemId)
                                else player.Events
                }
                do! DbService.updatePlayer updatedPlayer
                do! sendFollowUpMessage ctx $"Sold {item.Type} {item.Name} for {item.Price}! Current Balance: {updatedPlayer.Bank}"
            })
    })

let handleStoreEvents (_ : DiscordClient) (event : ComponentInteractionCreateEventArgs) =
    let ctx = DiscordEventContext event :> IDiscordContext
    let id = ctx.GetInteractionId()
    let itemId = int <| id.Split("-").[1]
    match id with
    | id when id.StartsWith("Buy") -> handleBuyItem ctx itemId
    | id when id.StartsWith("Sell") -> handleSell ctx itemId
    | _ ->
        task {
            let builder = DiscordInteractionResponseBuilder()
            builder.IsEphemeral <- true
            builder.Content <- $"Incorrect Action identifier {id}"
            do! ctx.Respond InteractionResponseType.ChannelMessageWithSource builder |> Async.AwaitTask
        }

type Store() =
    inherit ApplicationCommandModule ()

    let enforceChannel (ctx : IDiscordContext) (storeFn : IDiscordContext -> Task) =
        match ctx.GetChannel().Id with
        | id when id = GuildEnvironment.channelArmory -> storeFn ctx
        | _ ->
            task {
                let msg = $"You must go to <#{GuildEnvironment.channelArmory}> channel to buy or sell weapons"
                do! Messaging.sendSimpleResponse ctx msg
            }

    [<SlashCommand("buy-hack", "Purchase a hack attack you can use to earn GoodBoyTokenz")>]
    member _.BuyHack (ctx : InteractionContext) = enforceChannel (DiscordInteractionContext(ctx)) (buy ItemType.Hack)

    [<SlashCommand("buy-shield", "Purchase a hack shield so you can protect your GoodBoyTokenz")>]
    member this.BuyShield (ctx : InteractionContext) = enforceChannel (DiscordInteractionContext(ctx)) (buy ItemType.Shield)

    [<SlashCommand("sell-hack", "Sell a hack for GoodBoyTokenz")>]
    member this.SellHack (ctx : InteractionContext) = enforceChannel (DiscordInteractionContext(ctx)) (sell ItemType.Hack)

    [<SlashCommand("sell-shield", "Sell a shield for GoodBoyTokenz")>]
    member this.SellShield (ctx : InteractionContext) = enforceChannel (DiscordInteractionContext(ctx)) (sell ItemType.Shield)

